// Copyright (C) 2008-2009 - INRIA - Michael Baudin
// Copyright (C) 2010 - DIGITEO - Michael Baudin
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

function [order,msg] = assert_comparecomplex ( a , b , tol )
  // Compare complex numbers with a tolerance.
  // 
  // Calling Sequence
  //   order = assert_comparecomplex ( a , b , tol )
  //   [order,msg] = assert_comparecomplex ( a , b , tol )
  // 
  // Parameters
  //   a, b : the values to be compared
  //   tol : a 2x1 matrix, the tolerance
  //   tol(1) : the relative tolerance 
  //   tol(2) : the absolute tolerance
  //   order : a 1x1 matrix of floating point integers, the order. Returns order=0 is a is close to b, order=-1 if a < b, order=+1 if a > b.
  //   msg : a 1x1 matrix of strings, a message representing what happened during the comparison process.
  // 
  // Description
  // Compare first by real parts, then by imaginary parts.
  // Takes into account numerical accuracy issues, by using 
  // a mixed relative and absolute tolerance criteria.
  //
  // This algorithm is designed to be used into sorting 
  // algorithms.
  // It allows to take into account for the portability issues
  // related to the outputs of functions producing 
  // matrix of complex doubles.
  // If this algorithm is plugged into a sorting function, 
  // it allows to consistently produce a sorted matrix, 
  // where the order can be independent of the operating system,
  // the compiler or other forms of issues modifying the 
  // order (but not the values).
  // 
  // We state that a is close to b depending on the following algorithm.
  // First, we compare the real parts and use the condition :
  //   <programlisting>
  //     cond1 = ( abs(ar-br) <= tol(1) * max(abs(ar),abs(br)) ) + tol(2)
  //   </programlisting>
  // If this is true, we compare the imaginary parts and use the condition
  //   <programlisting>
  //     cond2 = ( abs(ai-bi) <= tol(1) * max(abs(ai),abs(bi)) ) + tol(2)
  //   </programlisting>
  // In the case where both cond1 and cond2 are true, 
  // then a and b are "close" and we return order=0.
  // If cond1 is true, but cond2 is false, we return order=-1 if 
  // (ai < bi) and order=1 if not.
  // If cond1 is false, then we return order=-1 if (ar<br) and order=1
  // if not.
  // 
  // Examples
  //   // Compare real values
  //   assert_comparecomplex ( 1 , -1 , [%eps 0] ) // 1
  //   [order,msg] = assert_comparecomplex ( 1 , -1 , [%eps 0] )
  //   assert_comparecomplex ( -1 , 1 , [%eps 0] ) // -1
  //   assert_comparecomplex ( 1 , 1 , [%eps 0] ) // 0
  //
  //   // Compare complex values #1
  //   assert_comparecomplex ( 1+2*%i , 1+3*%i , [%eps 0] ) // -1
  //   assert_comparecomplex ( 1+3*%i , 1+2*%i , [%eps 0] ) // 1
  //   assert_comparecomplex ( 1+2*%i , 1+2*%i , [%eps 0] ) // 0
  //
  //   // Compare complex values #2
  //   tol = [%eps 0]
  //   assert_comparecomplex ( 1+%i , -1+%i , tol ) // 1
  //   assert_comparecomplex ( -1+%i , 1+%i , tol ) // -1
  //   assert_comparecomplex ( 1+%i , 1+%i , tol ) // 0
  //   [order,msg] = assert_comparecomplex ( 1+%i , 1+%i , tol )
  //
  //   // Compare with tolerances : equality cases
  //   assert_comparecomplex ( 1.2345+%i , 1.2346+%i , [%eps 1.e-3] ) // 0
  //   assert_comparecomplex ( 1.2345+%i , 1.2346+%i , [1.e12*%eps 0] ) // 0
  //   assert_comparecomplex ( 1+1.2345*%i , 1+1.2347*%i , [%eps 1.e-3] ) // 0
  //   assert_comparecomplex ( 1+1.2345*%i , 1+1.2347*%i , [1.e12*%eps 0] ) // 0
  //
  //   // Compare more realistic data
  //  x = [
  //    -0.123452 - 0.123454 * %i
  //    -0.123451 + 0.123453 * %i
  //     0.123458 - 0.123459 * %i
  //     0.123456 + 0.123457 * %i
  //  ];
  //  // Consider less than 4 significant digits
  //  tol = [1.e-4 0];
  //  for i = 1 : size(x,"*")-1
  //    order = assert_comparecomplex ( x(i) , x(i+1) , tol );
  //    mprintf("compare(x(%d),x(%d))=%d\n",i,i+1,order)
  //  end
  //
  // // Compare data from bug #415
  //  x = [
  //    -1.9914145
  //    -1.895889
  //    -1.6923826
  //    -1.4815461
  //    -1.1302576
  //    -0.5652256 - 0.0655080 * %i
  //    -0.5652256 + 0.0655080 * %i
  //    0.3354023 - 0.1602902 * %i
  //    0.3354023 + 0.1602902 * %i
  //     1.3468911
  //     1.5040136
  //     1.846668
  //     1.9736772
  //     1.9798866
  //  ];
  //  // Consider less than 4 significant digits
  //  tol = [1.e-5 0];
  //  for i = 1 : size(x,"*")-1
  //    [order,msg] = assert_comparecomplex ( x(i) , x(i+1) , tol );
  //    mprintf("compare(x(%d),x(%d))=%d (%s)\n",i,i+1,order,msg)
  //  end
  //
  // Authors
  // Michael Baudin, DIGITEO, 2009-2010 (to manage a comparison function)
  //
  // Bibliography
  // http://gitweb.scilab.org/?p=scilab.git;a=blob;f=scilab/modules/polynomials/tests/nonreg_tests/bug_415.tst;h=0c716a3bed0dfb72c831972d19dbb0814dffde2b;hb=HEAD
  // http://gitweb.scilab.org/?p=scilab.git;a=blob_plain;f=scilab/modules/cacsd/tests/nonreg_tests/bug_68.tst;h=920d091d089b61bf961ea9e888b4d7d469942a14;hb=4ce3d4109dd752fce5f763be71ea639e09a12630

  msg = ""
  ar = real(a)
  br = real(b)
  areequal = ( abs ( ar - br ) <= tol(1) * max ( abs(ar) , abs(br) ) ) + tol(2)
  if ( areequal ) then
    msg = msg + " ar~br."
    // Tie on the real part: compare imaginary parts
    ai = imag(a)
    bi = imag(b)
    areequal = ( abs ( ai - bi ) <= tol(1) * max ( abs(ai) , abs(bi) ) ) + tol(2)
    if ( areequal ) then
      // Tie on imaginary parts too: two numbers are "equal"
      msg = msg + " ai~bi: numerical equality."
      order = 0
    elseif ( ai < bi ) then
      msg = msg + " ai<bi: a<b."
      order = -1
    else
      msg = msg + " ai>bi: a>b."
      order = 1
    end
  elseif ( ar < br ) then
    msg = msg + " ar<br: a<b."
    order = -1
  else
    msg = msg + " ar>br: a>b."
    order = 1
  end
endfunction

